import copy
import glob
import os
import pytest
import requests
import uuid
import zipfile

from le_utils.constants import licenses, content_kinds, exercises, roles
from ricecooker.__init__ import __version__
from ricecooker.classes.files import AudioFile, DocumentFile, EPubFile, HTMLZipFile, ThumbnailFile, SlideImageFile, SubtitleFile, VideoFile
from ricecooker.classes.files import _ExerciseImageFile, _ExerciseBase64ImageFile, _ExerciseGraphieFile
from ricecooker.classes.nodes import AudioNode, ChannelNode, DocumentNode, ExerciseNode, HTML5AppNode, SlideshowNode, TopicNode, VideoNode
from ricecooker.classes.questions import InputQuestion, SingleSelectQuestion



# GLOBAL TEST SETUP/TEARDOWN UTILS
################################################################################

def pytest_sessionfinish(session, exitstatus):
    """
    Cleanup testcontent/generated/ directory after each test run is finished.
    """
    generated_path = os.path.join("tests", "testcontent", "generated")
    for path in glob.glob(generated_path + os.path.sep + '*'):
        os.remove(path)


# CHANNEL FIXTURES
################################################################################

@pytest.fixture
def domain_namespace():
    return "testing.learningequality.org"

@pytest.fixture
def channel_source_id():
    return "channel-id"

@pytest.fixture
def channel_domain_namespace(domain_namespace):
    return uuid.uuid5(uuid.NAMESPACE_DNS, domain_namespace)

@pytest.fixture
def channel_node_id(channel_domain_namespace, channel_source_id):
    return uuid.uuid5(channel_domain_namespace, channel_source_id)

@pytest.fixture
def channel_content_id(channel_domain_namespace, channel_node_id):
    return uuid.uuid5(channel_domain_namespace, channel_node_id.hex)

@pytest.fixture
def channel_data(channel_node_id, channel_content_id, domain_namespace, channel_source_id):
    return {
        "id": channel_node_id.hex,
        "name": "Channel",
        "thumbnail": None,
        "description": "Channel description",
        "license": None,
        "source_domain": domain_namespace,
        "source_id": channel_source_id,
        "language": "en",
        "ricecooker_version": __version__,
    }

@pytest.fixture
def channel(domain_namespace, channel_source_id, channel_data):
    channel = ChannelNode(
        channel_source_id,
        domain_namespace,
        title=channel_data['name'],
        description=channel_data['description'],
        language=channel_data['language']
    )
    return channel

@pytest.fixture
def invalid_channel(channel_source_id, domain_namespace):
    channel = ChannelNode(
        channel_source_id,
        domain_namespace,
        title='Invalid Channel'
    )
    channel.source_id = None
    return channel




# ID, ARGS, AND KWARGS FIXTURE HELPERS
################################################################################

@pytest.fixture
def base_data(channel_domain_namespace, title):
    """
    The dictionary returned by this function resembles outpout of `to_dict` method.
    """
    return {
        "kind": None,
        "title": title,
        "description": "Description",
        "author": "Author",
        "source_domain": channel_domain_namespace.hex,
        "files" : [],
        "tags": [],
        "questions": [],
        "extra_fields": "{}",  # because Ricecookr uses `json.dumps` for this field
        "license": None,
        "copyright_holder": "",
        "license_description": None,
        "aggregator": "",           # New in ricecooker 0.6.20
        "provider": "",             # New in ricecooker 0.6.20
    }


def genrate_random_ids(channel_domain_namespace, channel_node_id):
    """
    Helper function to ensure all ContentNodes in test channel have unique `source_id`s.
    """
    source_id = uuid.uuid4().hex
    content_id = uuid.uuid5(channel_domain_namespace, source_id)
    node_id = uuid.uuid5(channel_node_id, content_id.hex)
    ids_dict = dict(
        source_id=source_id,
        content_id=content_id.hex,
        node_id=node_id.hex,
    )
    return ids_dict




# TOPIC FIXTURES
################################################################################

def get_topic_node_args(node_data):
    """
    Returns (source_id, title) from node_data dictionary.
    """
    node_data = copy.deepcopy(node_data)
    source_id = node_data.pop('source_id')
    title = node_data.pop('title')
    license = node_data.pop('license')
    return source_id, title

def get_topic_node_kwargs_data(node_data):
    """
    Returns all keywords data other than source_id, title, and license.
    """
    node_data = copy.deepcopy(node_data)
    del node_data['source_id']
    del node_data['title']
    # the following attributes will appear in `to_dict` method, but we don't need
    # to pass them in when creating a TopicNode
    del node_data['content_id']
    del node_data['node_id']
    del node_data['kind']
    del node_data['source_domain']
    del node_data['questions']
    del node_data['license']
    del node_data['license_description']
    del node_data['copyright_holder']
    return node_data


@pytest.fixture
def title():
    return "Title"


@pytest.fixture
def topic_data(base_data, channel_domain_namespace, channel_node_id):
    topic_data = copy.deepcopy(base_data)
    ids_dict = genrate_random_ids(channel_domain_namespace, channel_node_id)
    topic_data.update(ids_dict)
    topic_data.update({ "kind": content_kinds.TOPIC,
                        "extra_fields": {} })
    return topic_data

@pytest.fixture
def topic(channel, title, topic_data):
    args_data = get_topic_node_args(topic_data)
    topic_kwargs = get_topic_node_kwargs_data(topic_data)
    topic = TopicNode(*args_data, **topic_kwargs)
    channel.add_child(topic)
    return topic


# CONTENT NODE FIXTURES
################################################################################

@pytest.fixture
def contentnode_base_data(base_data):
    """
    Shared data for all ContentNode fixtures.
    """
    data = copy.deepcopy(base_data)
    data.update({ "license": licenses.CC_BY,
                  "copyright_holder": "Copyright Holder",
                  "license_description": None,
                  "role": roles.LEARNER})
    return data


def get_content_node_args(node_data):
    """
    Returns (source_id, title, license) from node_data dictionary.
    """
    node_data = copy.deepcopy(node_data)
    source_id = node_data.pop('source_id')
    title = node_data.pop('title')
    license = node_data.pop('license')
    return source_id, title, license


def get_content_node_kwargs(node_data):
    """
    Returns all keywords data other than source_id, title, and license.
    """
    node_data = copy.deepcopy(node_data)
    del node_data['source_id']
    del node_data['title']
    del node_data['license']
    # below are vars from internal representation
    del node_data['content_id']
    del node_data['node_id']
    del node_data['kind']
    del node_data['source_domain']
    del node_data['questions']
    node_data['extra_fields'] = {}
    return node_data


@pytest.fixture
def base_file_path():
    return "test/file/path"

@pytest.fixture
def contentnode_invalid_license(video):
    video = copy.deepcopy(video)
    video.license = None
    return video

@pytest.fixture
def contentnode_invalid_files(video):
    video = copy.deepcopy(video)
    video.files = []
    return video

@pytest.fixture
def contentnode_no_source_id(title):
    topic = TopicNode('some source id', title)
    topic.source_id = None
    return topic





# VIDEO FIXTURES
################################################################################

@pytest.fixture
def video_file():    # uses same file as test_videos.low_res_video fixture
    source_url = "https://archive.org/download/vd_is_for_everybody/vd_is_for_everybody_512kb.mp4"
    local_path = os.path.join("tests", "testcontent", "downloaded", "low_res_video.mp4")
    download_fixture_file(source_url, local_path)
    assert os.path.exists(local_path)
    return VideoFile(local_path)

@pytest.fixture
def video_filename():
    return '897d83a2e5389d454d37feb574587516.mp4'

@pytest.fixture
def subtitle_file():
    local_path = os.path.join("tests", "testcontent", "generated", "testsubtitles.vtt")
    if not os.path.exists(local_path):
        with open(local_path, 'wb') as subtitlefile:
            subtitlefile.write(b'WEBVTT\n')
            subtitlefile.write(b'\n')
            subtitlefile.write(b'00:01.000 --> 00:04.250\n')
            subtitlefile.write(b'Testing subtitles\n')
    return SubtitleFile(local_path, language='en')

@pytest.fixture
def subtitle_filename():
    return '19faefeb0b8b8289923dc0c1c5adb7e5.vtt'


@pytest.fixture
def video_data(contentnode_base_data, channel_domain_namespace, channel_node_id):
    video_data = copy.deepcopy(contentnode_base_data)
    ids_dict = genrate_random_ids(channel_domain_namespace, channel_node_id)
    video_data.update(ids_dict)
    video_data.update({ "kind": content_kinds.VIDEO })
    return video_data

@pytest.fixture
def video(video_file, video_data, channel):
    args_data = get_content_node_args(video_data)
    contentnode_kwargs = get_content_node_kwargs(video_data)
    video = VideoNode(*args_data, **contentnode_kwargs)
    video.add_file(video_file)
    channel.add_child(video)
    video_data['files'].append(video_file)  # save it so we can compare later
    return video

@pytest.fixture
def video_invalid_files(video_data, document_file):
    args_data = get_content_node_args(video_data)
    contentnode_kwargs = get_content_node_kwargs(video_data)
    contentnode_kwargs['files'] = []  # clear files becuse added one above
    video = VideoNode(*args_data, **contentnode_kwargs)
    video.add_file(document_file)
    return video

@pytest.fixture
def invalid_video_file():
    local_path = os.path.join("tests", "testcontent", "generated", "invalid_video.mp4")
    if not os.path.exists(local_path):
        with open(local_path, 'wb') as f:
            f.write(b'this is an invalid video file')
    return DocumentFile(local_path)

@pytest.fixture
def youtube_video_dict():
    """
    A short 17 sec video that won't slow down tests too much.
    """
    return {"youtube_id": "C0DPdy98e4c"}

@pytest.fixture
def youtube_video_with_subs_dict():
    """
    A test video that has subtitles in several languages.
    """
    return {
        "youtube_id": "USq6DX7byoY",
        "subtitles_langs": ["nl", "en", "en-GB", "fr", "el", "hu", "it", "pt", "ro", "es"]
    }

# AUDIO FIXTURES
################################################################################

@pytest.fixture
def audio_file():
    source_url = "https://ia800103.us.archive.org/9/items/cd_prince_prince/" \
                 "disc1/02.%20Prince%20-%201999%20%28Edit%29_sample.mp3"
    local_path = os.path.join("tests", "testcontent", "downloaded", "testaudio.mp3")
    download_fixture_file(source_url, local_path)
    assert os.path.exists(local_path)
    return AudioFile(local_path)

@pytest.fixture
def audio_filename():
    return 'c335e8044ecf583c690d5d8c65d68627.mp3'


@pytest.fixture
def audio_data(contentnode_base_data, audio_file, channel_domain_namespace, channel_node_id):
    audio_data = copy.deepcopy(contentnode_base_data)
    ids_dict = genrate_random_ids(channel_domain_namespace, channel_node_id)
    audio_data.update(ids_dict)
    audio_data.update({ "kind": content_kinds.AUDIO })
    return audio_data

@pytest.fixture
def audio(audio_file, audio_data, channel):
    args_data = get_content_node_args(audio_data)
    contentnode_kwargs = get_content_node_kwargs(audio_data)
    audio = AudioNode(*args_data, **contentnode_kwargs)
    audio.add_file(audio_file)
    channel.add_child(audio)
    audio_data['files'].append(audio_file)  # save it so we can compare later
    return audio

@pytest.fixture
def audio_invalid_files(audio_data, document_file):
    args_data = get_content_node_args(audio_data)
    contentnode_kwargs = get_content_node_kwargs(audio_data)
    contentnode_kwargs['files'] = []  # clear files because added one above
    audio = AudioNode(*args_data, **contentnode_kwargs)
    audio.add_file(document_file)
    return audio

@pytest.fixture
def invalid_audio_file():
    local_path = os.path.join("tests", "testcontent", "generated", "invalid_audio.mp3")
    if not os.path.exists(local_path):
        with open(local_path, 'wb') as f:
            f.write(b'invalid MP3')
    return DocumentFile(local_path)

# DOCUMENT FIXTURES
################################################################################

@pytest.fixture
def document_file():
    source_url = "https://ia802506.us.archive.org/8/items/generalmanual_000075878/generalmanual_000075878.pdf"
    local_path = os.path.join("tests", "testcontent", "downloaded", "testdocument.pdf")
    download_fixture_file(source_url, local_path)
    assert os.path.exists(local_path)
    return DocumentFile(local_path)

@pytest.fixture
def document_filename():
    return 'b976c31a7ab68a97f12541d661245238.pdf'

@pytest.fixture
def document_data(contentnode_base_data, document_file, channel_domain_namespace, channel_node_id):
    document_data = copy.deepcopy(contentnode_base_data)
    ids_dict = genrate_random_ids(channel_domain_namespace, channel_node_id)
    document_data.update(ids_dict)
    document_data.update({ "kind": content_kinds.DOCUMENT })
    return document_data

@pytest.fixture
def document(document_file, document_data, channel):
    args_data = get_content_node_args(document_data)
    contentnode_kwargs = get_content_node_kwargs(document_data)
    document = DocumentNode(*args_data, **contentnode_kwargs)
    document.add_file(document_file)
    channel.add_child(document)
    document_data['files'].append(document_file)  # save it so we can compare later
    return document

@pytest.fixture
def document_invalid_files(document_data, audio_file):
    args_data = get_content_node_args(document_data)
    contentnode_kwargs = get_content_node_kwargs(document_data)
    contentnode_kwargs['files'] = []  # clear files becuse added one above
    document = DocumentNode(*args_data, **contentnode_kwargs)
    document.add_file(audio_file)
    return document


@pytest.fixture
def epub_file():
    path = os.path.join("tests", "testcontent", "samples", "testdocument.epub")
    assert os.path.exists(path)
    return EPubFile(path)

@pytest.fixture
def epub_filename():
    return '5f91b55a7648206343b609cae692e08c.epub'


@pytest.fixture
def invalid_document_file():
    local_path = os.path.join("tests", "testcontent", "generated", "invalid_document.pdf")
    if not os.path.exists(local_path):
        with open(local_path, 'wb') as f:
            f.write(b'invalid PDF')
    return DocumentFile(local_path)

@pytest.fixture
def invalid_epub_file():
    local_path = os.path.join("tests", "testcontent", "generated", "invalid_document.epub")
    if not os.path.exists(local_path):
        with open(local_path, 'wb') as f:
            f.write(b'invalid ePub')
    return EPubFile(local_path)


# HTML FIXTURES
################################################################################

@pytest.fixture
def html_file():
    source_url = "https://studio.learningequality.org/content/storage/" \
                 "e/d/ed494d6547b603b8ff22095cf5f5b624.zip"
    local_path = os.path.join("tests", "testcontent", "downloaded", "testhtml.zip")
    download_fixture_file(source_url, local_path)
    assert os.path.exists(local_path)
    return HTMLZipFile(local_path)

@pytest.fixture
def html_filename():
    return 'ed494d6547b603b8ff22095cf5f5b624.zip'


@pytest.fixture
def html_data(contentnode_base_data, html_file, channel_domain_namespace, channel_node_id):
    html_data = copy.deepcopy(contentnode_base_data)
    ids_dict = genrate_random_ids(channel_domain_namespace, channel_node_id)
    html_data.update(ids_dict)
    html_data.update({ "kind": content_kinds.HTML5 })
    return html_data

@pytest.fixture
def html(html_file, html_data, channel):
    args_data = get_content_node_args(html_data)
    contentnode_kwargs = get_content_node_kwargs(html_data)
    html = HTML5AppNode(*args_data, **contentnode_kwargs)
    html.add_file(html_file)
    channel.add_child(html)
    html_data['files'].append(html_file)  # save it so we can compare later
    return html


@pytest.fixture
def html_invalid_files(html_data, document_file):
    """
    Invalid because adding a PDF document to a HTML5AppNode.
    """
    args_data = get_content_node_args(html_data)
    contentnode_kwargs = get_content_node_kwargs(html_data)
    contentnode_kwargs['files'] = []  # clear files becuse added one above
    html = HTML5AppNode(*args_data, **contentnode_kwargs)
    html.add_file(document_file)
    return html


@pytest.fixture
def html_invalid_file():
    local_path = os.path.join("tests", "testcontent", "generated", "testinvalidhtml.zip")
    if not os.path.exists(local_path):
        with zipfile.ZipFile(local_path, 'w', zipfile.ZIP_DEFLATED) as archive:
            archive.writestr("notindex.html", '<div></div>')
    return HTMLZipFile(local_path)

@pytest.fixture
def html_invalid_zip(html_data, html_invalid_file):
    args_data = get_content_node_args(html_data)
    contentnode_kwargs = get_content_node_kwargs(html_data)
    contentnode_kwargs['files'] = []  # clear files because added one above
    html = HTML5AppNode(*args_data, **contentnode_kwargs)
    html.add_file(html_invalid_file)
    return html



# EXERCISE FIXTURES
################################################################################

@pytest.fixture
def exercise_question():
    return SingleSelectQuestion("question_1", "Question", "Answer", ["Answer"])

@pytest.fixture
def mastery_model():
    return {'mastery_model': exercises.M_OF_N, 'randomize': True, 'm': 1, 'n': 1}

@pytest.fixture
def exercise_data(contentnode_base_data, mastery_model, exercise_question, channel_domain_namespace, channel_node_id):
    exercise_data = copy.deepcopy(contentnode_base_data)
    ids_dict = genrate_random_ids(channel_domain_namespace, channel_node_id)
    exercise_data.update(ids_dict)
    exercise_data.update({ "kind": content_kinds.EXERCISE,
                           "questions":[],
                           "exercise_data": mastery_model})
    return exercise_data

@pytest.fixture
def exercise(exercise_question, exercise_data, channel):
    args_data = get_content_node_args(exercise_data)
    contentnode_kwargs = get_content_node_kwargs(exercise_data)
    del contentnode_kwargs['extra_fields']
    mastery_model_dict = contentnode_kwargs['exercise_data']
    exercise = ExerciseNode(*args_data, **contentnode_kwargs)
    exercise.add_question(exercise_question)
    channel.add_child(exercise)
    exercise_data['questions'] = [exercise_question]
    exercise_data['extra_fields'] = mastery_model_dict
    del exercise_data['exercise_data']
    return exercise

@pytest.fixture
def exercise_invalid_question(exercise):
    exercise = copy.deepcopy(exercise)
    exercise.add_question(InputQuestion("question_2", "Question 2", ["Answer"]))
    return exercise



# THUMBNAIL FILE FIXTURES
################################################################################

@pytest.fixture
def thumbnail_file():
    local_path = os.path.join("tests", "testcontent", "samples", "thumbnail.png")
    assert os.path.exists(local_path)
    return ThumbnailFile(local_path)

@pytest.fixture
def thumbnail_filename():
    return 'eb79354ddd5774bb3436f9a19c282bff.png'

@pytest.fixture
def fake_thumbnail_file():
    local_path = os.path.join("tests", "testcontent", "generated", "invalidimage.png")
    if not os.path.exists(local_path):
        with open(local_path, 'wb') as imgfile:
            imgfile.write(b'not_a_valid_PNG')
    return ThumbnailFile(local_path)



# EXERCISE IMAGES FIXTURES
################################################################################

@pytest.fixture
def exercise_image_file():
    return _ExerciseImageFile('tests/testcontent/exercises/no-wifi.png')

@pytest.fixture
def exercise_image_filename():
    return '599aa896313be22dea6c0257772a464e.png'


@pytest.fixture
def exercise_base64_image_file():
    with open('tests/testcontent/exercises/test_image_base64.data') as datafile:
        base64_data = datafile.read()
        return _ExerciseBase64ImageFile(base64_data)

@pytest.fixture
def exercise_base64_image_filename():
    return 'cd9635def904486701e7705ef29ece67.png'


@pytest.fixture
def exercise_graphie_file():
    return _ExerciseGraphieFile('tests/testcontent/exercises/eb3f3bf7c317408ee90995b5bcf4f3a59606aedd')

@pytest.fixture
def exercise_graphie_replacement_str():
    return 'eb3f3bf7c317408ee90995b5bcf4f3a59606aedd'

@pytest.fixture
def exercise_graphie_filename():
    return 'ea2269bb5cf487f8d883144b9c06fbc7.graphie'




# SLIDESHOW IMAGES FIXTURES
################################################################################

@pytest.fixture
def slideshow_files():
    fake_files = []
    for i in range(0,10):
        filename = 'tests/testcontent/generated/slide' + str(i) + '.jpg'
        if not os.path.exists(filename):
            with open(filename, 'w') as f:
                f.write('jpgdatawouldgohere' + str(i))
        fake_files.append(
            SlideImageFile(filename, caption='slide ' + str(i))
        )
    return fake_files

@pytest.fixture
def slideshow_data(contentnode_base_data, slideshow_files, channel_domain_namespace, channel_node_id):
    slideshow_data = copy.deepcopy(contentnode_base_data)
    ids_dict = genrate_random_ids(channel_domain_namespace, channel_node_id)
    slideshow_data.update(ids_dict)
    slideshow_data.update({ "kind": content_kinds.SLIDESHOW })
    # TODO setup expected extra_fields['slideshow_data']
    return slideshow_data

@pytest.fixture
def slideshow(slideshow_files, slideshow_data, channel):
    args_data = get_content_node_args(slideshow_data)
    contentnode_kwargs = get_content_node_kwargs(slideshow_data)
    del contentnode_kwargs['extra_fields']
    slideshow = SlideshowNode(*args_data, **contentnode_kwargs)
    for slideshow_file in slideshow_files:
        slideshow.add_file(slideshow_file)
    channel.add_child(slideshow)
    slideshow_data['files'] = slideshow_files   # save it so we can compare later
    return slideshow



# FIXTURE DOWNLOADING UTILS
################################################################################

def download_fixture_file(source_url, local_path):
    """
    Download fixture file `source_url` to `local_path` if not present already.
    """
    if os.path.exists(local_path):
        return
    with open(local_path, 'wb') as f:
        response = requests.get(source_url, stream=True)
        for chunk in response.iter_content(chunk_size=1048576):
            f.write(chunk)
        f.flush()
        f.close()





import os
import PIL
import pytest

from le_utils.constants import licenses
from ricecooker import config
from ricecooker.classes.files import AudioFile, DocumentFile, EPubFile, HTMLZipFile, ThumbnailFile, TiledThumbnailFile, SlideImageFile, SubtitleFile, VideoFile
from ricecooker.classes.nodes import AudioNode, ChannelNode, DocumentNode, ExerciseNode, HTML5AppNode, SlideshowNode, TopicNode, VideoNode

from test_tree import thumbnail_path, thumbnail_path_jpg
from test_videos import low_res_video
from test_videos import _clear_ricecookerfilecache


SHOW_THUMBS = False     # set to True to show outputs when running tests locally


THUMBNAIL_URL = 'https://raw.githubusercontent.com/learningequality/ricecooker/master/tests/testcontent/samples/thumbnail.png'

class TestThumbnailSetting(object):

    def setup_method(self, test_method):
        """
        Called before each test method executes.
        """
        _clear_ricecookerfilecache()
        config.FAILED_FILES = []

    def get_video_node(self, path, thumbnail=None):
        video_file = VideoFile(path, language='en')
        video_node = VideoNode('vid-src-id', "Video", licenses.PUBLIC_DOMAIN, thumbnail=thumbnail)
        video_node.add_file(video_file)
        return video_node

    def check_correct_thumbnail(self, node):
        expected_thumbnail_filename = 'eb79354ddd5774bb3436f9a19c282bff.png'
        thumbnail_files = [f for f in node.files if isinstance(f, ThumbnailFile)]
        assert len(thumbnail_files) == 1, 'multiple thumbnails found'
        thumbnail_file = thumbnail_files[0]
        thumbnail_filename = thumbnail_file.get_filename()
        assert thumbnail_filename == expected_thumbnail_filename, "Wrong thumbnail"

    def assert_failed_thumbnail(self, node):
        thumbnail_files = [f for f in node.files if isinstance(f, ThumbnailFile)]
        assert len(thumbnail_files) == 1, 'multiple thumbnails found'
        thumbnail_file = thumbnail_files[0]
        assert thumbnail_file.filename == None, 'filename should be None'
        failed_files = config.FAILED_FILES
        # for ff in failed_files:
        #     print(ff, ff.path, ff.error)
        assert len(failed_files) == 1, 'multiple failed files found'
        failed_file = failed_files[0]
        assert failed_file.error, 'must have error set'
        assert thumbnail_file == failed_file, 'bad thumbnail file not found in config.FAILED_FILES'


    # HAPPY PATHS
    ############################################################################

    def test_set_png_thumbnail_from_local_path(self, low_res_video, thumbnail_path):
        video_node = self.get_video_node(path=low_res_video.name, thumbnail=thumbnail_path)
        video_node.validate()
        _ = video_node.process_files()
        self.check_correct_thumbnail(video_node)

    def test_set_jpg_thumbnail_from_local_path(self, low_res_video, thumbnail_path_jpg):
        video_node = self.get_video_node(path=low_res_video.name, thumbnail=thumbnail_path_jpg)
        video_node.validate()
        _ = video_node.process_files()
        expected_thumbnail_filename = 'd7ab03e4263fc374737d96ac2da156c1.jpg'
        thumbnail_files = [f for f in video_node.files if isinstance(f, ThumbnailFile)]
        assert len(thumbnail_files) == 1, 'multiple thumbnails found'
        thumbnail_file = thumbnail_files[0]
        thumbnail_filename = thumbnail_file.get_filename()
        assert thumbnail_filename == expected_thumbnail_filename, "Wrong thumbnail"

    def test_set_thumbnail_from_url(self, low_res_video):
        video_node = self.get_video_node(path=low_res_video.name, thumbnail=THUMBNAIL_URL)
        video_node.validate()
        _ = video_node.process_files()
        self.check_correct_thumbnail(video_node)

    def test_set_thumbnail_from_url_with_querystring(self, low_res_video):
        url = THUMBNAIL_URL + '?querystringkey=querystringvalue'
        video_node = self.get_video_node(path=low_res_video.name, thumbnail=url)
        video_node.validate()
        _ = video_node.process_files()
        self.check_correct_thumbnail(video_node)

    def test_set_thumbnail_from_ThumbnailFile(self, low_res_video, thumbnail_path):
        thumbnail_file = ThumbnailFile(thumbnail_path)
        video_node = self.get_video_node(path=low_res_video.name, thumbnail=thumbnail_file)
        video_node.validate()
        _ = video_node.process_files()
        self.check_correct_thumbnail(video_node)

    def test_add_ThumbnailFile(self, low_res_video, thumbnail_path):
        video_node = self.get_video_node(path=low_res_video.name, thumbnail=None)
        thumbnail_file = ThumbnailFile(thumbnail_path)
        video_node.add_file(thumbnail_file)
        video_node.validate()
        _ = video_node.process_files()
        self.check_correct_thumbnail(video_node)


    # ERROR PATHS
    ############################################################################

    def test_set_thumbnail_from_non_existent_path(self, low_res_video):
        non_existent_path = 'does/not/exist.png'
        video_node = self.get_video_node(path=low_res_video.name, thumbnail=non_existent_path)
        video_node.validate()
        _ = video_node.process_files()
        self.assert_failed_thumbnail(video_node)

    def test_set_thumbnail_from_bad_path(self, low_res_video, fake_thumbnail_file):
        """
        File path exists, but is not a valid PNG file.
        """
        video_node = self.get_video_node(path=low_res_video.name, thumbnail=fake_thumbnail_file)
        video_node.validate()
        _ = video_node.process_files()
        self.assert_failed_thumbnail(video_node)



class TestThumbnailGeneration(object):

    def setup_method(self, test_method):
        """
        Called before each test method executes.
        """
        _clear_ricecookerfilecache()
        config.FAILED_FILES = []
        config.THUMBNAILS = False

    def check_has_thumbnail(self, node):
        thumbnail_files = [f for f in node.files if isinstance(f, ThumbnailFile) or isinstance(f, TiledThumbnailFile)]
        assert len(thumbnail_files) == 1, 'expected single thumbnail'
        thumbnail_file = thumbnail_files[0]
        thumbnail_filename = thumbnail_file.get_filename()
        thumbnail_path = config.get_storage_path(thumbnail_filename)
        assert os.path.exists(thumbnail_path)
        img = PIL.Image.open(thumbnail_path)
        img.verify()
        if SHOW_THUMBS:
            img = PIL.Image.open(thumbnail_path)
            img.show()

    def assert_failed_thumbnail(self, node):
        thumbnail_files = [f for f in node.files if isinstance(f, ThumbnailFile)]
        assert len(thumbnail_files) == 1, 'multiple thumbnails found'
        thumbnail_file = thumbnail_files[0]
        assert thumbnail_file.filename == None, 'filename should be None'
        failed_files = config.FAILED_FILES
        # for ff in failed_files:
        #     print(ff, ff.path, ff.error)
        assert len(failed_files) == 1, 'multiple failed files found'
        failed_file = failed_files[0]
        assert failed_file.error, 'must have error set'
        assert thumbnail_file == failed_file, 'bad thumbnail file not found in config.FAILED_FILES'


    # HAPPY PATHS
    ############################################################################

    def test_generate_thumbnail_from_pdf(self, document_file):
        node = DocumentNode('doc-src-id', "Document", licenses.PUBLIC_DOMAIN, thumbnail=None)
        node.add_file(document_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        assert len(filenames) == 2, 'expected two filenames'
        self.check_has_thumbnail(node)

    def test_generate_thumbnail_from_epub(self, epub_file):
        node = DocumentNode('doc-src-id', "Document", licenses.PUBLIC_DOMAIN, thumbnail=None)
        node.add_file(epub_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        assert len(filenames) == 2, 'expected two filenames'
        self.check_has_thumbnail(node)

    def test_generate_thumbnail_from_html(self, html_file):
        node = HTML5AppNode('html-src-id', "HTML5 App", licenses.PUBLIC_DOMAIN, thumbnail=None)
        node.add_file(html_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        assert len(filenames) == 2, 'expected two filenames'
        self.check_has_thumbnail(node)

    def test_generate_thumbnail_from_video(self, video_file):
        node = VideoNode('vid-src-id', "Video", licenses.PUBLIC_DOMAIN, thumbnail=None)
        node.add_file(video_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        assert len(filenames) == 2, 'expected two filenames'
        self.check_has_thumbnail(node)

    def test_generate_thumbnail_from_audio(self, audio_file):
        node = AudioNode('audio-src-id', "Audio", licenses.PUBLIC_DOMAIN, thumbnail=None)
        node.add_file(audio_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        assert len(filenames) == 2, 'expected two filenames'
        self.check_has_thumbnail(node)

    def test_generate_tiled_thumbnail(self, document, html, video, audio):
        topic = TopicNode('test-topic', 'Topic')
        topic.add_child(document)
        topic.add_child(html)
        topic.add_child(video)
        topic.add_child(audio)
        config.THUMBNAILS = True
        for child in topic.children:   # must process children before topic node
            child.process_files()
        filenames = topic.process_files()
        assert len(filenames) == 1, 'expected one filename'
        self.check_has_thumbnail(topic)


    # ERROR PATHS
    ############################################################################

    def test_non_existent_pdf_fails(self):
        node = DocumentNode('doc-src-id', "Document", licenses.PUBLIC_DOMAIN, thumbnail=None)
        non_existent_path = 'does/not/exist.pdf'
        document_file = DocumentFile(non_existent_path, language='en')
        node.add_file(document_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        assert filenames == [None], 'expected one None (the non existent pdf)'
        assert len(config.FAILED_FILES) == 1, 'expected one failed file'

    def test_invalid_pdf_fails(self, invalid_document_file):
        node = DocumentNode('doc-src-id', "Document", licenses.PUBLIC_DOMAIN, thumbnail=None)
        node.add_file(invalid_document_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        # assert filenames == [None], 'expected one None filename (the broken pdf)'
        assert len(config.FAILED_FILES) == 1, 'expected one failed file'


    def test_non_existent_epub_fails(self):
        node = DocumentNode('doc-src-id', "Document", licenses.PUBLIC_DOMAIN, thumbnail=None)
        non_existent_path = 'does/not/exist.epub'
        document_file = DocumentFile(non_existent_path, language='en')
        node.add_file(document_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        assert filenames == [None], 'expected one None (the non existent epub)'
        assert len(config.FAILED_FILES) == 1, 'expected one failed file'

    def test_invalid_epub_fails(self, invalid_epub_file):
        node = DocumentNode('doc-src-id', "Document", licenses.PUBLIC_DOMAIN, thumbnail=None)
        node.add_file(invalid_epub_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        # assert filenames == [None], 'expected one None filename (the broken epub)'  # TODO: implement epub deep validation
        assert len(config.FAILED_FILES) == 1, 'expected one failed file'


    def test_non_existent_htmlzip_fails(self):
        node = HTML5AppNode('doc-src-id', "Document", licenses.PUBLIC_DOMAIN, thumbnail=None)
        non_existent_path = 'does/not/exist.zip'
        html_file = HTMLZipFile(non_existent_path, language='en')
        node.add_file(html_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        assert filenames == [None], 'expected one None filename (the broken zip)'
        assert len(config.FAILED_FILES) == 1, 'expected one failed file'

    def test_invalid_htmlzip_fails(self, html_invalid_file):
        node = DocumentNode('doc-src-id', "Document", licenses.PUBLIC_DOMAIN, thumbnail=None)
        node.add_file(html_invalid_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        assert filenames == [None], 'expected one None filename (the broken html)'
        assert len(config.FAILED_FILES) == 1, 'expected one failed file'


    def test_non_existent_mp3_fails(self):
        node = AudioNode('audio-src-id', "Document", licenses.PUBLIC_DOMAIN, thumbnail=None)
        non_existent_path = 'does/not/exist.mp3'
        document_file = AudioFile(non_existent_path, language='en')
        node.add_file(document_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        assert filenames == [None], 'expected one None (the non existent mp3)'
        assert len(config.FAILED_FILES) == 1, 'expected one failed file'

    def test_invalid_mp3_fails(self, invalid_audio_file):
        node = AudioNode('audio-src-id', "Document", licenses.PUBLIC_DOMAIN, thumbnail=None)
        node.add_file(invalid_audio_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        # assert filenames == [None], 'expected one None filename (the broken mp3)'   # TODO: implement mp3 deep validation
        # assert len(config.FAILED_FILES) == 1, 'expected one failed file'


    def test_non_existent_mp4_fails(self):
        node = VideoNode('video-src-id', "Video", licenses.PUBLIC_DOMAIN, thumbnail=None)
        non_existent_path = 'does/not/exist.mp4'
        document_file = VideoFile(non_existent_path, language='en')
        node.add_file(document_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        assert filenames == [None], 'expected one None (the non existent mp4)'
        assert len(config.FAILED_FILES) == 1, 'expected one failed file'

    def test_invalid_mp4_fails(self, invalid_video_file):
        node = VideoNode('video-src-id', "Document", licenses.PUBLIC_DOMAIN, thumbnail=None)
        node.add_file(invalid_video_file)
        config.THUMBNAILS = True
        filenames = node.process_files()
        # assert filenames == [None], 'expected one None filename (the broken mp4)'   # TODO: implement deep video validation
        # assert len(config.FAILED_FILES) == 1, 'expected one failed file'



import ntpath
import os
from pathlib import Path


def dir_exists(filepath):
    file_ = Path(filepath)
    return file_.is_dir()


def file_exists(filepath):
    my_file = Path(filepath)
    return my_file.is_file()


def get_name_from_url(url):
    """
    get the filename from a url 
    url = http://abc.com/xyz.txt
    get_name_from_url(url) -> xyz.txt
    """

    head, tail = ntpath.split(url)
    params_index = tail.find("&")
    if params_index != -1:
        tail = tail[:params_index]
    params_index = tail.find("?")
    if params_index != -1:
        tail = tail[:params_index]

    basename = ntpath.basename(url)
    params_b_index = basename.find("&")
    if params_b_index != -1:
        basename = basename[:params_b_index]
    return tail or basename


def get_name_from_url_no_ext(url):
    """
    get the filename without the extension name from a url 
    url = http://abc.com/xyz.txt
    get_name_from_url(url) -> xyz
    """
    path = get_name_from_url(url)
    return os.path.splitext(path)[0]


def build_path(levels):
    """
    make a linear directory structure from a list of path levels names
    levels = ["chefdir", "trees", "test"]
    builds ./chefdir/trees/test/
    """
    path = os.path.join(*levels)
    if not dir_exists(path):
        os.makedirs(path)
    return path



#!/usr/bin/env python
import argparse
import copy
import csv
from datetime import datetime
import dictdiffer
import json
import os
import requests

from ricecooker.config import LOGGER


# CONFIG CONSTANTS for data directories
################################################################################
STUDIO_CREDENTIALS='credentials/studio.json'
CHEFDATA_DIR = 'chefdata'
STUDIO_TREES_DIR = os.path.join(CHEFDATA_DIR, 'studiotrees')
if not os.path.exists(STUDIO_TREES_DIR):
    os.makedirs(STUDIO_TREES_DIR)
CORRECTIONS_DIR = os.path.join(CHEFDATA_DIR, 'corrections')
if not os.path.exists(CORRECTIONS_DIR):
    os.makedirs(CORRECTIONS_DIR)



# CORRECTIONS STRUCTURE v0.2
################################################################################

ACTION_KEY = 'Action'
NODE_ID_KEY = 'Node ID'
CONTENT_ID_KEY = 'Content ID'
PATH_KEY = 'Path'
CONTENT_KIND_KEY = 'Content Kind'
OLD_TITLE_KEY = 'Old Title'
NEW_TITLE_KEY = 'New Title'
OLD_DESCR_KEY = 'Old Description'
NEW_DESCR_KEY = 'New Description'
OLD_TAGS_KEY = 'Old Tags'
NEW_TAGS_KEY = 'New Tags'
OLD_COPYRIGHT_HOLDER_KEY = 'Old Copyright Holder'
NEW_COPYRIGHT_HOLDER_KEY = 'New Copyright Holder'
OLD_AUTHOR_KEY = 'Old Author'
NEW_AUTHOR_KEY = 'New Author'

CORRECTIONS_HEADER = [
    ACTION_KEY,
    NODE_ID_KEY,
    CONTENT_ID_KEY,
    PATH_KEY,
    CONTENT_KIND_KEY,
    OLD_TITLE_KEY,
    NEW_TITLE_KEY,
    OLD_DESCR_KEY,
    NEW_DESCR_KEY,
    OLD_TAGS_KEY,
    NEW_TAGS_KEY,
    OLD_COPYRIGHT_HOLDER_KEY,
    NEW_COPYRIGHT_HOLDER_KEY,
    OLD_AUTHOR_KEY,
    NEW_AUTHOR_KEY,
]


# What columns to export metadata to...
TARGET_COLUMNS = {
    'title': [OLD_TITLE_KEY, NEW_TITLE_KEY],
    'description': [OLD_DESCR_KEY, NEW_DESCR_KEY],
    'tags': [OLD_TAGS_KEY, NEW_TAGS_KEY],
    'copyright_holder': [OLD_COPYRIGHT_HOLDER_KEY, NEW_COPYRIGHT_HOLDER_KEY],
    'author': [OLD_AUTHOR_KEY, NEW_AUTHOR_KEY],
}

# default_keys = ['node_id', 'content_id'] # 'studio_id', 'source_id']
default_export = ['title', 'description', 'tags', 'copyright_holder', 'author']







# Studio Tree Local Cache queries
################################################################################

def get_channel_tree(api, channel_id, suffix='', update=True):
    """
    Downloads the entire main tree of a Studio channel to a local json file.
    """
    filename = os.path.join(STUDIO_TREES_DIR, channel_id + suffix + '.json')
    if os.path.exists(filename) and not update:
        print('  Loading cached tree for channel_id=', channel_id, 'from', filename)
        channel_tree = json.load(open(filename, 'r'))
        return channel_tree
    else:
        print('  Downloading tree for channel_id=', channel_id, ' and saving to', filename)
        root_studio_id = api.get_channel_root_studio_id(channel_id)
        # next step takes long since recursively making O(n) API calls!
        channel_tree = api.get_tree_for_studio_id(root_studio_id)
        json.dump(channel_tree, open(filename, 'w'), indent=4, ensure_ascii=False, sort_keys=True)
        return channel_tree



def print_channel_tree(channel_tree):
    """
    Print tree structure.
    """
    def print_tree(subtree, indent=''):
        kind = subtree.get("kind", 'topic')  # topic default to handle channel root
        if kind == "exercise":
            print(indent, subtree['title'],
                          'kind=', subtree['kind'],
                          len(subtree['assessment_items']), 'questions',
                          len(subtree['files']), 'files')
        else:
            print(indent, subtree['title'],
                          'kind=', subtree['kind'],
                          len(subtree['files']), 'files')
        for child in subtree['children']:
            print_tree(child, indent=indent+'    ')
    print_tree(channel_tree)



# CORECTIONS EXPORT
################################################################################

class CorretionsCsvFileExporter(object):

    def __init__(self, csvfilepath='corrections-export.csv', exportattrs=default_export):
        self.csvfilepath = csvfilepath
        self.exportattrs = exportattrs

    def download_channel_tree(self, api, channel_id):
        """
        Downloads a complete studio channel_tree from the Studio API.
        """
        channel_tree = get_channel_tree(api, channel_id, suffix='-export')
        return channel_tree


    # Export CSV metadata from external corrections
    ############################################################################

    def export_channel_tree_as_corrections_csv(self, channel_tree):
        """
        Create rows in corrections.csv from a Studio channel, specified based on
        node_id and content_id.
        """
        file_path = self.csvfilepath
        if os.path.exists(file_path):
            print('Overwriting previous export', file_path)
        with open(file_path, 'w') as csv_file:
            csvwriter = csv.DictWriter(csv_file, CORRECTIONS_HEADER)
            csvwriter.writeheader()

        def _write_subtree(path_tuple, subtree, is_root=False):
            # print('    '*len(path_tuple) + '  - ', subtree['title'])
            kind = subtree['kind']

            # TOPIC ############################################################
            if kind == 'topic':
                if is_root:
                    self.write_topic_row_from_studio_dict(path_tuple, subtree, is_root=is_root)
                    for child in subtree['children']:
                        _write_subtree(path_tuple, child)
                else:
                    self.write_topic_row_from_studio_dict(path_tuple, subtree)
                    for child in subtree['children']:
                        _write_subtree(path_tuple+[subtree['title']], child)

            # CONTENT NODES ####################################################
            elif kind in ['video', 'audio', 'document', 'html5']:
                self.write_content_row_from_studio_dict(path_tuple, subtree)

            # EXERCISE NODES ###################################################
            # elif kind == 'exercise':
            #     content_id = subtree['content_id']
            #     self.write_exercice_row_from_studio_dict(path_tuple, subtree, content_id)
            #     for question_dict in subtree['assessment_items']:
            #         self.write_question_row_from_question_dict(source_id, question_dict)
            else:
                print('>>>>> skipping node', subtree['title'])

        path_tuple = []
        _write_subtree(path_tuple, channel_tree, is_root=True)


    def write_common_row_attributes_from_studio_dict(self, row, studio_dict):
        # 1. IDENTIFIERS
        row[NODE_ID_KEY] = studio_dict['node_id']
        row[CONTENT_ID_KEY] = studio_dict['content_id']
        # PATH_KEY is set in specific function
        row[CONTENT_KIND_KEY] = studio_dict['kind']

        # 2. METADATA
        for exportattr in self.exportattrs:
            target_cols = TARGET_COLUMNS[exportattr]
            for target_col in target_cols:
                if exportattr == 'tags':
                    tags = studio_dict['tags']
                    tags_semicolon_separated = ';'.join(tags)
                    row[target_col] = tags_semicolon_separated
                else:
                    row[target_col] = studio_dict[exportattr]

    def write_topic_row_from_studio_dict(self, path_tuple, studio_dict, is_root=False):
        if is_root:
            return
        print('Generating corrections-export.csv rows for path_tuple ', path_tuple, studio_dict['title'])
        file_path = self.csvfilepath
        with open(file_path, 'a') as csv_file:
            csvwriter = csv.DictWriter(csv_file, CORRECTIONS_HEADER)
            title = studio_dict['title']
            path_with_self = '/'.join(path_tuple+[title])
            topic_row = {}
            self.write_common_row_attributes_from_studio_dict(topic_row, studio_dict)
            # WRITE TOPIC ROW
            topic_row[PATH_KEY] = path_with_self
            csvwriter.writerow(topic_row)

    def write_content_row_from_studio_dict(self, path_tuple, studio_dict):
        file_path = self.csvfilepath
        with open(file_path, 'a') as csv_file:
            csvwriter = csv.DictWriter(csv_file, CORRECTIONS_HEADER)
            row = {}
            self.write_common_row_attributes_from_studio_dict(row, studio_dict)
            title = studio_dict['title']
            row[PATH_KEY] = '/'.join(path_tuple+[title])
            # WRITE ROW
            csvwriter.writerow(row)








# CSV CORRECTIONS LOADERS
################################################################################

def save_gsheet_to_local_csv(gsheet_id, gid, csvfilepath='corrections-import.csv'):
    GSHEETS_BASE = 'https://docs.google.com/spreadsheets/d/'
    SHEET_CSV_URL = GSHEETS_BASE + gsheet_id + '/export?format=csv&gid=' + gid
    print(SHEET_CSV_URL)
    response = requests.get(SHEET_CSV_URL)
    csv_data = response.content.decode('utf-8')
    with open(csvfilepath, 'w') as csvfile:
        csvfile.write(csv_data)
        print('Succesfully saved ' + csvfilepath)
    return csvfilepath


def _clean_dict(row):
    """
    Transform empty strings values of dict `row` to None.
    """
    row_cleaned = {}
    for key, val in row.items():
        if val is None or val == '':
            row_cleaned[key] = None
        else:
            row_cleaned[key] = val.strip()
    return row_cleaned


def load_corrections_from_csv(csvfilepath):
    csv_path = csvfilepath     # download_structure_csv()
    struct_list = []
    with open(csv_path, 'r') as csvfile:
        reader = csv.DictReader(csvfile, fieldnames=CORRECTIONS_HEADER)
        next(reader)  # Skip Headers row
        for row in reader:
            clean_row = _clean_dict(row)
            struct_list.append(clean_row)
    return struct_list


def get_csv_corrections(csvfilepath):
    """
    Return a GROUP BY `corrkind` dictionary of rows from the CSV file.
    """
    modifications = []
    deletions = []
    rows = load_corrections_from_csv(csvfilepath)
    for i, row in enumerate(rows):
        if row[ACTION_KEY] == '' or row[ACTION_KEY] == None:
            print('Skipping no-action row', i+1)
        elif row[ACTION_KEY] == 'modify':
            modifications.append(row)
        elif row[ACTION_KEY] == 'delete':
            deletions.append(row)
        else:
            print('Uknown Action', row[ACTION_KEY])
    return {
        'modifications': modifications,
        'deletions': deletions,
    }


def get_corrections_by_node_id(csvfilepath, modifyattrs):
    """
    Convert CSV to internal representaiton of corrections as dicts by node_id.
    """
    corrections_by_node_id = {
        'nodes_modified': {},
        'nodes_added': {},
        'nodes_deleted': {},
        'nodes_moved': {},
    }
    csv_corrections = get_csv_corrections(csvfilepath)  # CSV rows GROUP BY corrkind
    #
    # Modifications
    for row in csv_corrections['modifications']:
        node_id = row[NODE_ID_KEY]
        # print('Found MODIFY row of CSV  for node_id', node_id)
        #
        # find all modified attributes
        attributes = {}
        for attr in modifyattrs:
            # print('Found MODIFY', attr, 'in row of CSV for node_id', node_id)
            old_key = TARGET_COLUMNS[attr][0]
            new_key = TARGET_COLUMNS[attr][1]
            if row[new_key] == row[old_key]:   # skip if the same
                continue
            else:
                attributes[attr] = {
                    'changed': True,
                    'value': row[new_key],
                    'old_value': row[old_key],
                }
        # prepare modifications_dict
        modifications_dict = {
            'attributes': attributes,
        }
        # add to to corrections_by_node_id
        corrections_by_node_id['nodes_modified'][node_id] = modifications_dict
    #
    # Deletions
    for row in csv_corrections['deletions']:
        node_id = row[NODE_ID_KEY]
        # print('Found DELETE row in CSV for node_id', node_id)
        corrections_by_node_id['nodes_deleted'][node_id] = {'node_id':node_id}
    #
    # TODO: Additions
    # TODO: Moves
    datetimesuffix = datetime.now().strftime("%Y-%m-%d__%H%M")
    correctionspath = os.path.join(CORRECTIONS_DIR, 'imported-' + datetimesuffix + '.json')
    json.dump(corrections_by_node_id, open(correctionspath, 'w'), indent=4, ensure_ascii=False, sort_keys=True)
    #
    return correctionspath



# Tree querying API
################################################################################


def find_nodes_by_attr(subtree, attr, value):
    """
    Returns list of nodes in `subtree` that have attribute `attr` equal to `value`.
    """
    results = []
    if subtree[attr] == value:
        results.append(subtree)
    if 'children' in subtree:
        for child in subtree['children']:
            child_restuls = find_nodes_by_attr(child, attr, value)
            results.extend(child_restuls)
    return results

def find_nodes_by_content_id(subtree, content_id):
    return find_nodes_by_attr(subtree, 'content_id', content_id)

def find_nodes_by_node_id(subtree, node_id):
    return find_nodes_by_attr(subtree, 'node_id', node_id)

def find_nodes_by_original_source_node_id(subtree, original_source_node_id):
    return find_nodes_by_attr(subtree, 'original_source_node_id', original_source_node_id)

def unresolve_children(node):
    """
    Return copy of node with children = list of studio_id references instead of full data.
    """
    node =  copy.deepcopy(node)
    if 'children' in node:
        new_children = []
        for child in node['children']:
            new_children.append(child['id'])
        node['children'] = new_children
    return node






# SPECIAL REMAP NEEDED FOR ALDARYN CORRECTIONS
################################################################################

def remap_original_source_node_id_to_node_id(channel_tree, corrections_by_original_source_node_id):
    ALL_COORECTIONS_KINDS = ['nodes_modified', 'nodes_added', 'nodes_deleted', 'nodes_moved']
    corrections_by_node_id = {}
    for correction_kind in ALL_COORECTIONS_KINDS:
        if correction_kind in corrections_by_original_source_node_id:
            corrections_by_node_id[correction_kind] = {}
            corrections_dict = corrections_by_original_source_node_id[correction_kind]
            for original_source_node_id, correction in corrections_dict.items():
                results = find_nodes_by_original_source_node_id(channel_tree, original_source_node_id)
                assert results, 'no match found based on original_source_node_id search'
                assert len(results)==1, 'multiple matches found...'
                tree_node = results[0]
                node_id = tree_node['node_id'] 
                corrections_by_node_id[correction_kind][node_id] = correction
    return corrections_by_node_id






# CORRECTIONS API CALLS
################################################################################

def apply_modifications_for_node_id(api, channel_tree, node_id, modifications_dict):
    """
    Given a modification dict of the form,
        modifications_dict = {
            'attributes': {
                'title': {
                    'changed': (bool),
                    'value': (str),
                    'old_value': (str),
                },
                'files': ([{
                    'filename': (str),
                    'file_size': (int),
                    'preset': (str)
                }]),
                'assessment_items': ([AssessmentItem]),
                'tags': ([Tag]),
                ...
            }
        }
    this function will make obtain GET the current node data from Studio API,
    apply the modifications to the local json data, then PUT the data on Studio.
    """
    # print('MODIFYING node_id=', node_id)
    results = find_nodes_by_node_id(channel_tree, node_id)
    assert results, 'no match found based on node_id search'
    assert len(results)==1, 'multiple matches found...'
    tree_node = results[0]
    studio_id = tree_node['id']
    # node_before = unresolve_children(tree_node)
    node_before = api.get_contentnode(studio_id)
    # print('node_before', node_before)

    # PREPARE data for PUT request  (starting form copy of old)
    data = {}
    ATTRS_TO_COPY = ['kind', 'id', 'tags', 'prerequisite', 'parent']
    for attr in ATTRS_TO_COPY:
        data[attr] = node_before[attr]
    #
    # ADD new_values modified 
    modifications = modifications_dict['attributes']
    for attr, values_diff in modifications.items():
        if values_diff['changed']:
            current_value = node_before[attr]
            expected_old_value = values_diff['old_value']
            new_value = values_diff['value']
            if expected_old_value == new_value:   # skip if the same
                continue
            if current_value != expected_old_value:
                print('WARNING expected old value', expected_old_value, 'for', attr, 'but current node value is', current_value)
            # print('Changing current_value', current_value, 'for', attr, 'to new value', new_value)
            data[attr] = new_value
        else:
            print('Skipping attribute', attr, 'because key changed==False')

    # PUT
    print('PUT studio_id=', studio_id, 'node_id=', node_id)
    response_data = api.put_contentnode(data)

    # Check what changed
    node_after = api.get_contentnode(studio_id)
    diffs = list(dictdiffer.diff(node_before, node_after))
    print('  diff=', diffs)
    return response_data


def apply_deletion_for_node_id(api, channel_tree, channel_id, node_id, deletion_dict):
    results = find_nodes_by_node_id(channel_tree, node_id)
    assert results, 'no match found based on node_id search'
    assert len(results)==1, 'multiple matches found...'
    tree_node = results[0]
    studio_id = tree_node['id']
    # node_before = unresolve_children(tree_node)
    node_before = api.get_contentnode(studio_id)
    
    # PREPARE data for DLETE request
    data = {}
    data['id'] = node_before['id']

    # DELETE
    print('DELETE studio_id=', studio_id, 'node_id=', node_id)
    response_data = api.delete_contentnode(data, channel_id)

    # Check what changed
    node_after = api.get_contentnode(studio_id)
    diffs = list(dictdiffer.diff(node_before, node_after))
    print('  diff=', diffs)

    return response_data


def apply_corrections_by_node_id(api, channel_tree, channel_id, corrections_by_node_id):
    """
    Given a dict `corrections_by_node_id` of the form,
    {
        'nodes_modified': {
            '<node_id (str)>': { modification dict1 },
            '<node_id (str)>': { modification dict2 },
        }
        'nodes_added': {
            '<node_id (str)>': { 'new_parent': (str),  'attributes': {...}},
        },
        'nodes_deleted': {
            '<node_id (str)>': {'old_parent': (str), 'attributes': {...}},
        },
        'nodes_moved': {
            '<node_id (str)>': {'old_parent': (str), 'new_parent': (str), 'attributes': {...}},
        },
    }
    this function will make the appropriate Studio API calls to apply the patch.
    """
    LOGGER.debug('Applying corrections...')
    #
    # Modifications
    for node_id, modifications_dict in corrections_by_node_id['nodes_modified'].items():
        apply_modifications_for_node_id(api, channel_tree, node_id, modifications_dict)
    #
    # Deletions
    for node_id, deletion_dict in corrections_by_node_id['nodes_deleted'].items():
        apply_deletion_for_node_id(api, channel_tree, channel_id, node_id, deletion_dict)
    # TODO: Additions
    # TODO: Moves







from ricecooker.utils.libstudio import StudioApi

def get_studio_api(studio_creds=None):
    if studio_creds is None:
        if not os.path.exists(STUDIO_CREDENTIALS):
            print('ERROR: Studio credentials file', STUDIO_CREDENTIALS, 'not found')
            print("""Please create the file and put the following informaiton in it:
            {
              "token": "<your studio token>",
              "username": "<your studio username>",
              "password": "<your studio password>",
            }
            """)
            raise ValueError('Missing credentials')
        studio_creds = json.load(open(STUDIO_CREDENTIALS))
    #
    # Studio API client (note currently needs both session auth and token as well)
    api = StudioApi(
            token=studio_creds['token'],
            username=studio_creds['username'],
            password=studio_creds['password'],
            studio_url=studio_creds.get('studio_url', 'https://studio.learningequality.org')
    )
    return api


def export_corrections_csv(args):
    api = get_studio_api()
    channel_tree = get_channel_tree(api, args.channel_id, suffix='-export')
    print_channel_tree(channel_tree)
    csvexporter = CorretionsCsvFileExporter()
    csvexporter.export_channel_tree_as_corrections_csv(channel_tree)


def apply_corrections(args):
    # 1. LOAD Studio channel_tree (needed for lookups by node_id, content_id, etc.)
    api = get_studio_api()
    channel_tree = get_channel_tree(api, args.channel_id, suffix='-before')
    #
    # 2. IMPORT the corrections from the Spreadsheet
    csvfilepath = 'corrections-import.csv'
    save_gsheet_to_local_csv(args.gsheet_id, args.gid, csvfilepath=csvfilepath)
    #
    # 3. TRANSFORM corrections-import.csv to Studio detailed diff format
    modifyattrs = args.modifyattrs.split(',')   # using only selected attributes
    correctionspath = get_corrections_by_node_id(csvfilepath, modifyattrs)
    #
    # Special case: when export was performed on source channel, but we want to
    # apply the corrections to a cloned channel. In that cases, the `Node ID`
    # column in the CSV corresponds to the `original_source_node_id` attribute
    # of the nodes in the derivative channel so we must do a remapping:
    if args.primarykey == 'original_source_node_id':
        corrections_by_original_source_node_id = json.load(open(correctionspath))
        corrections_by_node_id = remap_original_source_node_id_to_node_id(channel_tree, corrections_by_original_source_node_id)
        json.dump(corrections_by_node_id, open(correctionspath, 'w'), indent=4, ensure_ascii=False, sort_keys=True)
        print('Finished original_source_node_id-->node_id lookup and remapping.')
    elif args.primarykey in ['content_id', 'studio_id']:
        raise NotImplementedError('Using content_id and studio_id not ready yet.')
    #
    # Early exit if running the `importonly` command
    if args.command == 'importonly':
        print('Corrections json file imported. See', correctionspath)
        return correctionspath
    #
    # 4. LOAD corrections.json (four lists of corrections organized by nod_id)
    corrections_by_node_id = json.load(open(correctionspath))
    #
    # 5. Apply the corrections
    apply_corrections_by_node_id(api, channel_tree, args.channel_id, corrections_by_node_id)
    #
    # 6. SAVE the Studio tree after corrections for review of what was changed
    channel_tree = get_channel_tree(api, args.channel_id, suffix='-after')



def correctionsmain():
    """
    Command line interface for applying bulk-edit corrections:
    """
    parser = argparse.ArgumentParser(description='Bulk channel edits via CSV/sheets.')
    parser.add_argument('command', help='One of export|importonly|apply',
                        choices=['export', 'importonly', 'apply'])
    parser.add_argument('channel_id', help='The studio Channel ID to edit')
    parser.add_argument('--primarykey', help='Which idendifier to use when looking up nodes',
                        choices=['node_id', 'content_id', 'original_source_node_id', 'studio_id'],
                        default='node_id')
    parser.add_argument('--gsheet_id', help='Google spreadsheets sheet ID (public)')
    parser.add_argument('--gid', help='The gid argument to indicate which sheet', default='0')
    parser.add_argument('--modifyattrs', help='Which attributes to modify',
                        default='title,description,author,copyright_holder')
    args = parser.parse_args()
    # print("in corrections.main with cliargs", args)
    if args.command == 'export':
        export_corrections_csv(args)
    elif args.command in ['importonly', 'apply']:
        apply_corrections(args)
    else:
        raise ValueError('Unrecognized command')


if __name__ == '__main__':
    correctionsmain()

